import h3.api.numpy_int as h3
import numpy as np

from h3ronpy.op import kring_distances_agg, kring_distances


def test_kring_distances_agg():
    h3indexes = np.array([h3.geo_to_h3(10.2, 45.5, 8), h3.geo_to_h3(10.3, 45.1, 8)], dtype=np.uint64)
    k_max = 4
    df = kring_distances_agg(h3indexes, k_max, aggregation_method='min')
    assert len(df) > 100
    assert df['k'].min() == 0
    assert df['k'].max() == k_max
    assert len(np.unique(df["h3index"])) == len(df)


def test_kring_distances():
    h3indexes = np.array([h3.geo_to_h3(10.2, 45.5, 8), h3.geo_to_h3(10.3, 45.1, 8)], dtype=np.uint64)
    k_max = 4
    k_min = 2
    df = kring_distances(h3indexes, 4, k_min=k_min)
    assert len(df) > 100
    assert df['ring_k'].min() == k_min
    assert df['ring_k'].max() == k_max
    assert 'h3index' in df
    assert 'ring_h3index' in df



try:
    import rasterio
    HAS_RASTERIO = True
except ImportError:
    # rasterio provides no pre-compiled windows wheels, so it is an
    # optional dependency to simplify the CI setup of the github actions workflows
    HAS_RASTERIO = False

import numpy as np
import pytest

from h3ronpy.raster import raster_to_dataframe

from . import TESTDATA_PATH

@pytest.mark.skipif(not HAS_RASTERIO, reason="requires rasterio")
def test_r_tiff():
    dataset = rasterio.open(TESTDATA_PATH / "r.tiff")
    band = dataset.read(1)
    df = raster_to_dataframe(band, dataset.transform, 8, nodata_value=0, compacted=True, geo=False)
    assert len(df) > 100
    assert df.dtypes["h3index"] == "uint64"
    assert df.dtypes["value"] == "uint8"


@pytest.mark.skipif(not HAS_RASTERIO, reason="requires rasterio")
def test_r_tiff_float32():
    dataset = rasterio.open(TESTDATA_PATH / "r.tiff")
    band = dataset.read(1).astype(np.float32)
    df = raster_to_dataframe(band, dataset.transform, 8, nodata_value=0.0, compacted=True, geo=False)
    assert len(df) > 100
    assert df.dtypes["h3index"] == "uint64"
    assert df.dtypes["value"] == "float32"



import typing

import numpy as np
import pandas as pd

from .h3ronpy import op as native_op


def kring_distances(h3indexes: np.ndarray, k_max: int, k_min: int = 0) -> pd.DataFrame:
    """
    Vectorized k-ring building

    Returns a dataframe with the columns `h3index` (the ring center), `ring_h3index` and `ring_k`.

    :param h3indexes:
    :param k_max:
    :param k_min:
    :return:
    """
    h3index, ring_h3index, ring_k = native_op.kring_distances(h3indexes, k_min, k_max)
    return pd.DataFrame({
        "h3index": h3index,
        "ring_h3index": ring_h3index,
        "ring_k": ring_k
    })


def kring_distances_agg_np(h3indexes: np.ndarray, k_max: int, k_min: int = 0, aggregation_method: str = 'min') -> \
        typing.Tuple[np.ndarray, np.ndarray]:
    """
    Vectorized k-ring building, with the k-values of the rings being aggregated to their `min` or
    `max` value for each cell.

    :param h3indexes:
    :param k_max:
    :param k_min:
    :param aggregation_method:
    :return:
    """
    return native_op.kring_distances_agg(h3indexes, k_min, k_max, aggregation_method)


def kring_distances_agg(h3indexes: np.ndarray, k_max: int, k_min: int = 0,
                        aggregation_method: str = 'min') -> pd.DataFrame:
    h3indexes_out, k_out = kring_distances_agg_np(h3indexes, k_max, k_min=k_min, aggregation_method=aggregation_method)
    return pd.DataFrame({"h3index": h3indexes_out, "k": k_out})



# install the dependencies needed for development and ci by
# collecting them from all relevant files

import subprocess
from pathlib import Path

import sys
import os

def install(packages, upgrade=False):
    pkg_manager = os.environ.get("PKG_MANAGER") or 'pip'
    cmd = [pkg_manager, "install"]
    if upgrade and pkg_manager == 'pip':
        cmd.append("--upgrade")
    if packages:
        subprocess.run(cmd + packages, stdout=sys.stdout, stderr=sys.stderr)


if __name__ == '__main__':
    install(["pip", "toml"], upgrade=True)  # always upgrade pip

    import toml  # import only after it has been installed

    directory = Path(__file__).parent
    packages = []

    pyproject_toml = toml.load(directory / "pyproject.toml")
    for pkg in pyproject_toml.get("build-system", {}).get("requires"):
        packages.append(pkg)

    pytest = pyproject_toml.get("tool", {}).get("pytest")
    if pytest is not None:
        pytest_package = "pytest"
        pytest_minversion = pytest.get("ini_options", {}).get("minversion")
        if pytest_minversion:
            packages.append(f"{pytest_package}>={pytest_minversion}")
        else:
            packages.append(f"{pytest_package}")

    for pkg in toml.load(directory / "Cargo.toml").get("package", {}).get("metadata", {}).get("maturin", {}).get(
            "requires-dist", []):
        packages.append(pkg)
    install(packages)



import geopandas as gpd
from h3ronpy.vector import geodataframe_to_h3


def test_geodataframe_to_h3():
    world = gpd.read_file(gpd.datasets.get_path('naturalearth_lowres'))
    africa = world[world["continent"] == "Africa"]
    df = geodataframe_to_h3(africa, 4)
    assert len(df) > 1000
    assert df.dtypes["h3index"] == "uint64"



import warnings
from typing import Union, List, Iterable

import geopandas as gpd
import h3.api.numpy_int as h3
import numpy as np
import pandas as pd
from shapely.geometry import Polygon

from . import H3_CRS


def h3indexes_to_geodataframe(
        h3indexes: Union[np.array, List[int], Iterable[int]]) -> gpd.GeoDataFrame:
    """
    convert a numpy-array with uint64 h3 indexes to a geodataframe
    """
    i_arr = np.asarray(h3indexes)
    return gpd.GeoDataFrame({
        "geometry": [Polygon(h3.h3_to_geo_boundary(h, geo_json=True)) for h in _array_nditer(i_arr)],
        "h3resolution": [h3.h3_get_resolution(h) for h in _array_nditer(i_arr)],
        "h3index": [h3.h3_to_string(h) for h in _array_nditer(i_arr)]
    }, crs=H3_CRS)


def h3index_column_to_geodataframe(df: pd.DataFrame, column_name: str = "h3index") -> gpd.GeoDataFrame:
    """
    convert a dataframe with a column containing h3indexes to a geodataframe

    :param df: input dataframe
    :param column_name: name of the column containing the h3 indexes
    :return: GeoDataFrame
    """
    warnings.warn("h3index_column_to_geodataframe has been deprecated in favor of dataframe_to_geodataframe",
                  DeprecationWarning)
    return dataframe_to_geodataframe(df, column_name=column_name)


def dataframe_to_geodataframe(df: pd.DataFrame, column_name: str = "h3index") -> gpd.GeoDataFrame:
    """
    convert a dataframe with a column containing h3indexes to a geodataframe

    :param df: input dataframe
    :param column_name: name of the column containing the h3 indexes
    :return: GeoDataFrame
    """
    return gpd.GeoDataFrame(df,
                            geometry=[Polygon(h3.h3_to_geo_boundary(h, geo_json=True)) for h in
                                      _array_nditer(df[column_name].to_numpy())],
                            crs=H3_CRS)


def _array_nditer(a: np.array):
    return np.nditer(a, flags=["zerosize_ok", ])



import pandas as pd

from h3ronpy.util import dataframe_to_geodataframe


def test_empty_to_geo():
    # https://github.com/nmandery/h3ron/issues/17
    df = pd.DataFrame({"h3index": []})
    gdf = dataframe_to_geodataframe(df)  # should not raise an ValueError.
    assert gdf.empty



from typing import Tuple, Generator

import geopandas as gpd
import math
import numpy as np
import pandas as pd

from .h3ronpy import vector


def geometries_to_h3_generator(geometries: np.array, ids: np.array, h3_resolution: int, do_compact: bool = False,
                               chunk_size: int = 1000) -> Generator[Tuple[np.array, np.array], None, None]:
    """
    Generator to convert shapely geometries and ids to two numpy arrays with h3indexes and the correlated Ids.
    Yields (ids, h3indexes)-tuples.

    This function is parallelized and uses the available CPUs by distributing the geometries to a thread pool.

    :param geometries: numpy-array with shapely geometries (in WGS84).
    :param ids: numpy array with ids. must be unit64
    :param h3_resolution: H3 resolution to use.
    :param do_compact: Compacts the h3index when this is set to `True`. Default is `False`
    :param chunk_size: Number of geometries to include in the yielded chunks
    :return: None
    """
    num_chunks = math.ceil(len(geometries) / chunk_size)
    for (chunk_geometries, chunk_ids) in zip(np.array_split(geometries, num_chunks),
                                             np.array_split(ids, num_chunks)):
        chunk_wkb_list = [item.wkb for item in chunk_geometries]

        (ids, h3indexes) = vector.wkbbytes_with_ids_to_h3(chunk_ids, chunk_wkb_list, h3_resolution, do_compact)
        yield ids, h3indexes


def geodataframe_to_h3(df: gpd.GeoDataFrame, h3_resolution: int, do_compact: bool = False, geometry_column="geometry",
                       index_column_name="to_h3_idx", chunk_size=1000):
    """
    convert the geometries of a geodataframe to h3 indexes.

    This function is parallelized and uses the available CPUs by distributing the geometries to a thread pool.

    The geometry column will be removed from the resulting dataframe. The input geometry must be in WGS84.

    :param df: The input geodataframe
    :param h3_resolution: H3 resolution to use.
    :param do_compact: Compacts the h3index when this is set to `True`. Default is `False`
    :param geometry_column: The name of the column containing the geometry. Defaults to `geometry`.
    :param index_column_name: The name for a temporary column used to join the H3 data to the input dataframe
    :param chunk_size:
    :return:
    """
    # add a column with a sequence to merge later with
    df.insert(loc=0, column=index_column_name, value=np.arange(len(df), dtype=np.uint64))
    dataframes = []
    for (ids, h3indexes) in geometries_to_h3_generator(df[geometry_column].to_numpy(), df[index_column_name].to_numpy(),
                                                       h3_resolution,
                                                       do_compact=do_compact, chunk_size=1000):
        dataframes.append(pd.DataFrame({
            index_column_name: ids,
            "h3index": h3indexes
        }))

    if not dataframes:
        return pd.DataFrame({})
    output_df = pd.DataFrame(df.drop(columns=geometry_column)) \
        .merge(pd.concat(dataframes), on=index_column_name)

    # remove the column used for the merge again
    del output_df[index_column_name]

    return output_df



"""
Conversion of raster numpy arrays to h3.

Resolution search modes
-----------------------

* "min_diff": chose the h3 resolution where the difference in the area of a pixel and the h3index is as small as possible.
* "smaller_than_pixel":  chose the h3 resolution where the area of the h3index is smaller than the area of a pixel.
"""

import geopandas as gp
import h3.api.numpy_int as h3
import numpy as np
from shapely.geometry import Polygon

import pandas as pd
from .h3ronpy import raster
from .util import _array_nditer
from . import H3_CRS

try:
    # affine library is used by rasterio
    import affine

    __HAS_AFFINE_LIB = True
except ImportError:
    __HAS_AFFINE_LIB = False


def _get_transform(t):
    if isinstance(t, raster.Transform):
        return t
    if __HAS_AFFINE_LIB:
        if isinstance(t, affine.Affine):
            return raster.Transform.from_rasterio([t.a, t.b, t.c, t.d, t.e, t.f])
    if type(t) in (list, tuple) and len(t) == 6:
        # probably native gdal
        return raster.Transform.from_gdal(t)
    raise ValueError("unsupported object for transform")


def nearest_h3_resolution(shape, transform, axis_order="yx", search_mode="min_diff"):
    """
    find the h3 resolution closed to the size of a pixel in an array
    of the given shape with the given transform

    :param shape: dimensions of the 2d array
    :param transform: the affine transformation
    :param axis_order: axis order of the 2d array. Either "xy" or "yx"
    :param search_mode: resolution search mode (see documentation of this module)
    :return:
    """
    return raster.nearest_h3_resolution(shape, _get_transform(transform), axis_order, search_mode)


def raster_to_dataframe(in_raster: np.array, transform, h3_resolution: int, nodata_value=None, axis_order: str = "yx",
                        compacted: bool = True, geo: bool = False):
    """
    convert a raster/array to a pandas dataframe containing H3 indexes

    This function is parallelized and uses the available CPUs by distributing tiles to a thread pool.

    The input geometry must be in WGS84.

    :param in_raster: input 2-d array
    :param transform:  the affine transformation
    :param nodata_value: the nodata value. For these cells of the array there will be no h3 indexes generated
    :param axis_order: axis order of the 2d array. Either "xy" or "yx"
    :param h3_resolution: target h3 resolution
    :param compacted: return compacted h3 indexes (see H3 docs)
    :param geo: return a geopandas geodataframe with geometries. increases the memory usage.
    :return: pandas dataframe or geodataframe
    """

    dtype = in_raster.dtype
    func = None
    if dtype == np.uint8:
        func = raster.raster_to_h3_u8
    elif dtype == np.int8:
        func = raster.raster_to_h3_i8
    elif dtype == np.uint16:
        func = raster.raster_to_h3_u16
    elif dtype == np.int16:
        func = raster.raster_to_h3_i16
    elif dtype == np.uint32:
        func = raster.raster_to_h3_u32
    elif dtype == np.int32:
        func = raster.raster_to_h3_i32
    elif dtype == np.uint64:
        func = raster.raster_to_h3_u64
    elif dtype == np.int64:
        func = raster.raster_to_h3_i64
    elif dtype == np.float32:
        func = raster.raster_to_h3_f32
    elif dtype == np.float64:
        func = raster.raster_to_h3_f64
    else:
        raise NotImplementedError(f"no raster_to_h3 implementation for dtype {dtype.name}")

    values, indexes = func(in_raster, _get_transform(transform), nodata_value, h3_resolution, axis_order, compacted)
    if geo:
        return gp.GeoDataFrame({
            "h3index": indexes,
            "value": values,
            "geometry": [Polygon(h3.h3_to_geo_boundary(h, geo_json=True)) for h in _array_nditer(indexes)],
        }, crs=H3_CRS)
    else:
        return pd.DataFrame({
            "h3index": indexes,
            "value": values
        })


def raster_to_geodataframe(*a, **kw):
    """
    convert to a geodataframe

    Uses the same parameters as array_to_dataframe
    """
    kw["geo"] = True
    return raster_to_dataframe(*a, **kw)
